@@ -25,14 +25,17 @@
 #include "eds-helper.h"
 #include "notify-message.h"
 #include "provider-template.h"
+#include "sync-network.h"
 
 #include <QtCore/QDebug>
 #include <QtCore/QTimer>
 
+#include <url-dispatcher.h>
+
 using namespace Accounts;
 
 
-#define DAEMON_SYNC_TIMEOUT     1000 * 6 // one minute
+#define DAEMON_SYNC_TIMEOUT     1000 * 60 // one minute
 #define SYNC_MONITOR_ICON_PATH  "/usr/share/icons/ubuntu-mobile/actions/scalable/reload.svg"
 
 SyncDaemon::SyncDaemon()
@@ -40,13 +43,18 @@ SyncDaemon::SyncDaemon()
       m_manager(0),
       m_eds(0),
       m_dbusAddaptor(0),
+      m_currentAccount(0),
       m_syncing(false),
-      m_aboutToQuit(false)
+      m_aboutToQuit(false),
+      m_firstClient(true)
 {
     m_provider = new ProviderTemplate();
     m_provider->load();
 
     m_syncQueue = new SyncQueue();
+    m_offlineQueue = new SyncQueue();
+    m_networkStatus = new SyncNetwork(this);
+    connect(m_networkStatus, SIGNAL(stateChanged(SyncNetwork::NetworkState)), SLOT(onOnlineStatusChanged(SyncNetwork::NetworkState)));
 
     m_timeout = new QTimer(this);
     m_timeout->setInterval(DAEMON_SYNC_TIMEOUT);
@@ -57,6 +65,10 @@ SyncDaemon::SyncDaemon()
 SyncDaemon::~SyncDaemon()
 {
     quit();
+    delete m_timeout;
+    delete m_syncQueue;
+    delete m_offlineQueue;
+    delete m_networkStatus;
 }
 
 void SyncDaemon::setupAccounts()
@@ -81,7 +93,7 @@ void SyncDaemon::setupAccounts()
 
 void SyncDaemon::setupTriggers()
 {
-    m_eds = new EdsHelper(this);
+    m_eds = new EdsHelper(this, "");
     connect(m_eds, &EdsHelper::dataChanged,
             this, &SyncDaemon::onDataChanged);
 }
@@ -100,8 +112,53 @@ void SyncDaemon::onDataChanged(const QString &serviceName, const QString &source
     }
 }
 
+void SyncDaemon::onClientAttached()
+{
+    if (m_firstClient) {
+        m_firstClient = false;
+        // accept eds changes
+        qDebug() << "First client connected, will auto-sync on next EDS change";
+        m_eds->setEnabled(true);
+     }
+}
+
+void SyncDaemon::onOnlineStatusChanged(SyncNetwork::NetworkState state)
+{
+    Q_EMIT isOnlineChanged(state != SyncNetwork::NetworkOffline);
+    if (state == SyncNetwork::NetworkOnline) {
+        qDebug() << "Device is online sync pending changes" << m_offlineQueue->count();
+        m_syncQueue->push(*m_offlineQueue);
+        m_offlineQueue->clear();
+        if (!m_syncing && !m_syncQueue->isEmpty()) {
+            qDebug() << "Will sync in" << DAEMON_SYNC_TIMEOUT / 1000 << "secs;";
+            m_syncing = true;
+            m_timeout->start();
+        } else {
+            qDebug() << "No change to sync";
+        }
+    } else if (state == SyncNetwork::NetworkOffline) {
+        qDebug() << "Device is offline cancel active syncs. There is a sync in progress?" << (m_currentAccount ? "Yes" : "No");
+        if (m_currentAccount) {
+            if (m_currentAccount->retrySync()) {
+                qDebug() << "Push sync to later sync";
+                m_offlineQueue->push(m_currentAccount, m_currentServiceName, false);
+            } else {
+                 qDebug() << "Do not try re-sync the account";
+            }
+            m_currentAccount->cancel(m_currentServiceName);
+        }
+        if (m_timeout->isActive()) {
+            m_timeout->stop();
+        }
+        continueSync();
+    }
+    // make accounts available or not based on online status
+    Q_EMIT accountsChanged();
+}
+
 void SyncDaemon::syncAll(const QString &serviceName, bool runNow)
 {
+    // if runNow is set we will sync all accounts
     Q_FOREACH(SyncAccount *acc, m_accounts.values()) {
         if (serviceName.isEmpty()) {
             sync(acc, QString(), runNow);
@@ -111,6 +168,45 @@ void SyncDaemon::syncAll(const QString &serviceName, bool runNow)
     }
 }
 
+/*
+ * This is a helper function used on contact migration it only works for contacts sync
+ */
+void SyncDaemon::syncAccount(quint32 accountId, const QString &service)
+{
+    Account *account = m_manager->account(accountId);
+
+    if (account) {
+        // fake a settings object
+        QSettings *contactSettings = new QSettings;
+        contactSettings->setValue("contacts/template", "WebDAV");
+        contactSettings->setValue("contacts/syncURL", "https://www.googleapis.com/.well-known/carddav");
+        contactSettings->setValue("contacts/uoa-service", "google-carddav");
+        contactSettings->setValue("contacts/sync-backend", "evolution-contacts");
+        contactSettings->setValue("contacts/sync-uri", "addressbook");
+
+        SyncAccount *acc = new SyncAccount(account, service, contactSettings, this);
+        if (!isOnline()) {
+            qWarning() << "Network is off ignore sync request";
+            Q_EMIT syncError(acc, service, "20017");
+            account->deleteLater();
+        } else {
+            acc->setRetrySync(false);
+            connect(acc, SIGNAL(syncStarted(QString, bool)),
+                         SLOT(onAccountSyncStarted(QString, bool)));
+            connect(acc, SIGNAL(syncFinished(QString, bool, QString, QString)),
+                         SLOT(onAccountSyncFinished(QString, bool, QString, QString)));
+            connect(acc, SIGNAL(syncError(QString,QString)),
+                         SLOT(onAccountSyncError(QString,QString)));
+            connect(acc, SIGNAL(syncError(QString,QString)),
+                    acc, SLOT(deleteLater()), Qt::QueuedConnection);
+            connect(acc, SIGNAL(syncFinished(QString,bool,QString,QString)),
+                    acc, SLOT(deleteLater()), Qt::QueuedConnection);
+            contactSettings->setParent(acc);
+            sync(acc, service, true);
+        }
+    }
+}
+
 void SyncDaemon::cancel(const QString &serviceName)
 {
     Q_FOREACH(SyncAccount *acc, m_accounts.values()) {
@@ -122,20 +218,28 @@ void SyncDaemon::cancel(const QString &serviceName)
     }
 }
 
-void SyncDaemon::sync(bool runNow)
-{
-    m_syncing = true;
-    if (runNow) {
-        m_timeout->stop();
-        continueSync();
-    } else {
-        // wait some time for new sync requests
-        m_timeout->start();
-    }
-}
-
 void SyncDaemon::continueSync()
 {
+    SyncJob job = m_syncQueue->popNext();
+    SyncNetwork::NetworkState netState = m_networkStatus->state();
+    bool continueSync = (netState == SyncNetwork::NetworkOnline) ||
+                        (netState != SyncNetwork::NetworkOffline && job.runOnPayedConnection());
+    if (!continueSync) {
+        qDebug() << "Device is offline we will skip the sync.";
+
+        Q_FOREACH(const SyncJob &j, m_syncQueue->jobs()) {
+            if (j.account() && j.account()->retrySync()) {
+                qDebug() << "Push account to later sync3";
+                m_offlineQueue->push(j);
+            }
+        }
+
+        m_syncQueue->clear();
+        syncFinishedImpl();
+        return;
+    }
+    m_syncing = true;
+
     // flush any change in EDS
     m_eds->flush();
 
@@ -143,16 +247,21 @@ void SyncDaemon::continueSync()
     m_eds->freezeNotify();
 
     // sync the next service on the queue
-    if (!m_aboutToQuit && !m_syncQueue->isEmpty()) {
-        m_currentServiceName = m_syncQueue->popNext(&m_currentAccount);
-        m_currentAccount->sync(m_currentServiceName);
+    if (!m_aboutToQuit && job.isValid()) {
+        m_currentServiceName = job.serviceName();
+        m_currentAccount = job.account();
     } else {
         m_currentAccount = 0;
-        m_currentServiceName.clear();
-        m_syncing = false;
+    }
+
+    if (m_currentAccount) {
+        // remove sync reqeust from offline queue
+        m_offlineQueue->remove(m_currentAccount, m_currentServiceName);
+        m_currentAccount->sync(m_currentServiceName);
+    } else {
+        syncFinishedImpl();
         // The sync has done, unblock notifications
         m_eds->unfreezeNotify();
-        Q_EMIT done();
     }
 }
 
@@ -169,17 +278,27 @@ bool SyncDaemon::registerService()
         }
 
         m_dbusAddaptor = new SyncDBus(connection, this);
-        if (!connection.registerObject(SYNCMONITOR_OBJECT_PATH, this))
-        {
+        if (!connection.registerObject(SYNCMONITOR_OBJECT_PATH, this)) {
             qWarning() << "Could not register object!" << SYNCMONITOR_OBJECT_PATH;
             delete m_dbusAddaptor;
             m_dbusAddaptor = 0;
             return false;
         }
+        connect(m_dbusAddaptor, SIGNAL(clientAttached(int)), SLOT(onClientAttached()));
     }
     return true;
 }
 
+void SyncDaemon::syncFinishedImpl()
+{
+    m_timeout->stop();
+    m_currentAccount = 0;
+    m_currentServiceName.clear();
+    m_syncing = false;
+    Q_EMIT done();
+    qDebug() << "All syncs finished";
+}
+
 void SyncDaemon::run()
 {
     setupAccounts();
@@ -189,9 +308,16 @@ void SyncDaemon::run()
     registerService();
 }
 
+bool SyncDaemon::isPending() const
+{
+    // there is a sync request on the buffer
+    return (m_syncQueue && (m_syncQueue->count() > 0));
+}
+
 bool SyncDaemon::isSyncing() const
 {
-    return m_syncing;
+    // the sync is happening right now
+    return (m_syncing && (m_currentAccount != 0));
 }
 
 QStringList SyncDaemon::availableServices() const
@@ -214,6 +340,11 @@ QStringList SyncDaemon::enabledServices() const
     return services.toList();
 }
 
+bool SyncDaemon::isOnline() const
+{
+    return m_networkStatus->state() != SyncNetwork::NetworkOffline;
+}
+
 void SyncDaemon::addAccount(const AccountId &accountId, bool startSync)
 {
     Account *acc = m_manager->account(accountId);
@@ -227,14 +358,14 @@ void SyncDaemon::addAccount(const AccountId &accountId, bool startSync)
         m_accounts.insert(accountId, syncAcc);
         connect(syncAcc, SIGNAL(syncStarted(QString, bool)),
                          SLOT(onAccountSyncStarted(QString, bool)));
-        connect(syncAcc, SIGNAL(syncFinished(QString, bool, QString)),
-                         SLOT(onAccountSyncFinished(QString, bool, QString)));
-        connect(syncAcc, SIGNAL(syncError(QString, int)),
-                         SLOT(onAccountSyncError(QString, int)));
+        connect(syncAcc, SIGNAL(syncFinished(QString, bool, QString, QString)),
+                         SLOT(onAccountSyncFinished(QString, bool, QString, QString)));
         connect(syncAcc, SIGNAL(enableChanged(QString, bool)),
                          SLOT(onAccountEnableChanged(QString, bool)));
         connect(syncAcc, SIGNAL(configured(QString)),
                          SLOT(onAccountConfigured(QString)), Qt::DirectConnection);
+        connect(syncAcc, SIGNAL(syncError(QString,QString)),
+                         SLOT(onAccountSyncError(QString, QString)));
         if (startSync) {
             sync(syncAcc, QString(), true);
         }
@@ -242,34 +373,55 @@ void SyncDaemon::addAccount(const AccountId &accountId, bool startSync)
     }
 }
 
+void SyncDaemon::sync(bool runNow)
+{
+    m_syncing = true;
+    if (runNow) {
+        m_timeout->stop();
+        continueSync();
+    } else {
+        // wait some time for new sync requests
+        m_timeout->start();
+    }
+}
+
 void SyncDaemon::sync(SyncAccount *syncAcc, const QString &serviceName, bool runNow)
 {
     qDebug() << "syn requested for account:" << syncAcc->displayName() << serviceName;
 
     // check if the service is already in the sync queue or is the current operation
     if (m_syncQueue->contains(syncAcc, serviceName) ||
-        (m_currentAccount == syncAcc) && (serviceName.isEmpty() || (serviceName == m_currentServiceName))) {
-        qDebug() << "Account aready in the queue";
+        ((m_currentAccount == syncAcc) && (serviceName.isEmpty() || (serviceName == m_currentServiceName))) ) {
+        qDebug() << "Account aready in the queue, ignore request;";
     } else {
-        qDebug() << "Pushed into queue";
-        m_syncQueue->push(syncAcc, serviceName);
+        qDebug() << "Pushed into queue with immediately sync?" << runNow << "Sync is running" << m_syncing;
+        m_syncQueue->push(syncAcc, serviceName, runNow);
         // if not syncing start a full sync
         if (!m_syncing) {
+            qDebug() << "Request sync";
             sync(runNow);
             Q_EMIT syncAboutToStart();
+            return;
         }
     }
+
+    // immediately request, force sync to start
+    if (runNow && !isSyncing()) {
+        sync(runNow);
+        Q_EMIT syncAboutToStart();
+    }
 }
 
 void SyncDaemon::cancel(SyncAccount *syncAcc, const QString &serviceName)
 {
-    NotifyMessage *notify = new NotifyMessage(true, this);
-    notify->show(_("Synchronization"),
-                  QString(_("Sync canceled: %1")).arg(syncAcc->displayName()),
-                 syncAcc->iconName(serviceName));
-
     m_syncQueue->remove(syncAcc, serviceName);
-    syncAcc->cancel(serviceName);
+    syncAcc->cancel();
+    if (m_currentAccount == syncAcc) {
+        qDebug() << "Current sync canceled";
+        m_currentAccount = 0;
+    } else if (m_syncQueue->isEmpty()) {
+        syncFinishedImpl();
+    }
     Q_EMIT syncError(syncAcc, serviceName, "canceled");
 }
 
@@ -278,29 +430,11 @@ void SyncDaemon::removeAccount(const AccountId &accountId)
     SyncAccount *syncAcc = m_accounts.take(accountId);
     if (syncAcc) {
         cancel(syncAcc);
-
-        NotifyMessage *notify = new NotifyMessage(true, this);
-        notify->setProperty("ACCOUNT", QVariant::fromValue<QObject*>(qobject_cast<QObject*>(syncAcc)));
-        connect(notify, SIGNAL(questionRejected()), SLOT(removeAccountSource()));
-        connect(notify, SIGNAL(messageClosed()), SLOT(destroyAccount()));
-        notify->askYesOrNo(_("Synchronization"),
-                           QString(_("Account %1 removed. Do you want to keep the account data?"))
-                                .arg(syncAcc->displayName()),
-                           SYNC_MONITOR_ICON_PATH);
-
+        m_eds->removeSource("", syncAcc->displayName());
     }
     Q_EMIT accountsChanged();
 }
 
-void SyncDaemon::removeAccountSource()
-{
-    QObject *sender = QObject::sender();
-    QObject *accObj = sender->property("ACCOUNT").value<QObject*>();
-    SyncAccount *acc = qobject_cast<SyncAccount*>(accObj);
-    Q_ASSERT(acc);
-    m_eds->removeSource("", acc->displayName());
-}
-
 void SyncDaemon::destroyAccount()
 {
     QObject *sender = QObject::sender();
@@ -309,69 +443,114 @@ void SyncDaemon::destroyAccount()
     acc->deleteLater();
 }
 
+void SyncDaemon::authenticateAccount(const SyncAccount *account, const QString &serviceName)
+{
+    NotifyMessage *notify = new NotifyMessage(true, this);
+    notify->setProperty("ACCOUNT", QVariant::fromValue<AccountId>(account->id()));
+    notify->setProperty("SERVICE", QVariant::fromValue<QString>(account->serviceId(serviceName)));
+    connect(notify, SIGNAL(questionAccepted()), SLOT(runAuthentication()));
+    notify->askYesOrNo(_("Synchronization"),
+                       QString(_("Your access key is not valid anymore. Do you want to re-authenticate it?.")),
+                       account->iconName(serviceName));
+
+}
+
+void SyncDaemon::runAuthentication()
+{
+    QObject *sender = QObject::sender();
+    AccountId accountId = sender->property("ACCOUNT").value<AccountId>();
+    QString serviceName = sender->property("SERVICE").value<QString>();
+
+    QString appCommand = QString("syncmonitorhelper:///authenticate?id=%1&service=%2").arg(accountId).arg(serviceName);
+    qDebug() << "Run" << appCommand;
+    url_dispatch_send(appCommand.toUtf8().constData(), NULL, NULL);
+}
+
 void SyncDaemon::onAccountSyncStarted(const QString &serviceName, bool firstSync)
 {
+    SyncAccount *acc = qobject_cast<SyncAccount*>(QObject::sender());
     if (firstSync) {
         NotifyMessage *notify = new NotifyMessage(true, this);
         notify->show(_("Synchronization"),
                      QString(_("Start sync:  %1 (%2)"))
-                         .arg(m_currentAccount->displayName())
+                         .arg(acc->displayName())
                          .arg(serviceName),
-                     m_currentAccount->iconName(serviceName));
+                     acc->iconName(serviceName));
     }
     m_syncElapsedTime.restart();
     qDebug() << QString("[%3] Start sync:  %1 (%2)")
-                .arg(m_currentAccount->displayName())
+                .arg(acc->displayName())
                 .arg(serviceName)
                 .arg(QDateTime::currentDateTime().toString(Qt::SystemLocaleShortDate));
-    Q_EMIT syncStarted(m_currentAccount, serviceName);
+    Q_EMIT syncStarted(acc, serviceName);
 }
 
-void SyncDaemon::onAccountSyncFinished(const QString &serviceName, const bool firstSync, const QString &status)
+void SyncDaemon::onAccountSyncError(const QString &serviceName, const QString &error)
 {
+    Q_EMIT syncError(qobject_cast<SyncAccount*>(QObject::sender()), serviceName, error);
+    onAccountSyncFinished(serviceName, false, error, "fast");
+}
+
+void SyncDaemon::onAccountSyncFinished(const QString &serviceName, const bool firstSync, const QString &status, const QString &mode)
+{
+    // error on that list will trigger a new sync
+    static QStringList whiteListStatus;
+
+    uint errorCode = status.toUInt();
+    SyncAccount *acc = qobject_cast<SyncAccount*>(QObject::sender());
     QString errorMessage = SyncAccount::statusDescription(status);
+
     if (firstSync && errorMessage.isEmpty()) {
         NotifyMessage *notify = new NotifyMessage(true, this);
         notify->show(_("Synchronization"),
                      QString(_("Sync done: %1 (%2)"))
-                         .arg(m_currentAccount->displayName())
+                         .arg(acc->displayName())
                          .arg(serviceName),
-                     m_currentAccount->iconName(serviceName));
-    } else if (!errorMessage.isEmpty()) {
-        NotifyMessage *notify = new NotifyMessage(true, this);
-        notify->show(_("Synchronization"),
-                     QString(_("Fail to sync %1 (%2).\n%3"))
-                         .arg(m_currentAccount->displayName())
-                         .arg(serviceName)
-                         .arg(errorMessage),
-                     m_currentAccount->iconName(serviceName));
+                     acc->iconName(serviceName));
     }
 
+    Q_EMIT syncFinished(acc, serviceName);
+
     qDebug() << QString("[%6] Sync done: %1 (%2) Status: %3 Error: %4 Duration: %5s")
-                .arg(m_currentAccount->displayName())
+                .arg(acc->displayName())
                 .arg(serviceName)
                 .arg(status)
                 .arg(errorMessage.isEmpty() ? "None" : errorMessage)
                 .arg((m_syncElapsedTime.elapsed() < 1000 ? 1  : m_syncElapsedTime.elapsed() / 1000))
                 .arg(QDateTime::currentDateTime().toString(Qt::SystemLocaleShortDate));
 
-    Q_EMIT syncFinished(m_currentAccount, serviceName);
 
-    // sync next account
-    continueSync();
-}
+    // populate white list erros
+    if (whiteListStatus.isEmpty()) {
+        // "error code from SyncEvolution access denied (remote, status 403): could not obtain OAuth2 token:
+        // this can happen if the network goes off during the sync, or syc started before the network stabilished
+        whiteListStatus << QStringLiteral("10403");
+        whiteListStatus << QStringLiteral("403");
 
-void SyncDaemon::onAccountSyncError(const QString &serviceName, int errorCode)
-{
-    NotifyMessage *notify = new NotifyMessage(true, this);
-    notify->show(_("Synchronization"),
-                 QString(_("Sync error account: %1, %2, %3"))
-                     .arg(m_currentAccount->displayName())
-                     .arg(serviceName)
-                     .arg(errorCode),
-                 m_currentAccount->iconName(serviceName));
+        // error code from SyncEvolution fatal error (local, status 10500): no sources active, check configuration"
+        // this is a bug on SyncEvolution sometimes it fail to read the correct address book
+        // FIXME: we should fix that on SyncEvolution
+        whiteListStatus << QStringLiteral("10500");
+    }
 
-    Q_EMIT syncError(m_currentAccount, serviceName, QString(errorCode));
+    // only re-sync if sync mode != "slow", to avoid sync loops
+    if ((acc->lastError() == 0) && !errorMessage.isEmpty() && whiteListStatus.contains(status)) {
+        // white list error retry the sync
+        m_syncQueue->push(acc, serviceName, false);
+    } else if (status.endsWith("403")){
+        authenticateAccount(acc, serviceName);
+        errorCode = 0;
+    } else if (!errorMessage.isEmpty()) {
+        NotifyMessage *notify = new NotifyMessage(true, this);
+        notify->show(_("Synchronization"),
+                     QString(_("Fail to sync %1 (%2).\n%3"))
+                         .arg(acc->displayName())
+                         .arg(serviceName)
+                         .arg(errorMessage),
+                     acc->iconName(serviceName));
+    }
+
+    acc->setLastError(errorCode);
     // sync next account
     continueSync();
 }
@@ -409,7 +588,16 @@ void SyncDaemon::quit()
 
     // cancel all sync operation
     while(m_syncQueue->count()) {
-        SyncAccount *acc = m_syncQueue->popNext();
+        SyncJob job = m_syncQueue->popNext();
+        SyncAccount *acc = job.account();
+        acc->cancel();
+        acc->wait();
+        delete acc;
+    }
+
+    while(m_offlineQueue->count()) {
+        SyncJob job = m_offlineQueue->popNext();
+        SyncAccount *acc = job.account();
         acc->cancel();
         acc->wait();
         delete acc;
@@ -419,4 +607,8 @@ void SyncDaemon::quit()
         delete m_manager;
         m_manager = 0;
     }
+
+    if (m_networkStatus) {
+        delete m_networkStatus;
+    }
 }
